#include "time.h"
#include <crtdbg.h>

#include "gamex-cleanmem.h"
#undef new
#undef delete
#undef malloc
#undef free

///#define DEBUG_OUTPUT_ALL_ALLOCS
#ifdef DEBUG_OUTPUT_ALL_ALLOCS
	static long totalAllocated = 0;
#	define DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
#	ifdef DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
#		include "windows.h"
#	endif
#endif

#ifdef CLEAN_MEM_DISABLE

	// if cleanMem was requested to be disabled,
	// do nothing but supply 2 empty functions:
	void BeginCheckingUnfreed(void) {}
	void DumpUnfreed(void) {}

#else // !CLEAN_MEM_DISABLE:

	// stuff for keeping track of allocations / deallocations
	struct ALLOC_INFO
	{
		unsigned long address;
		unsigned long size;
		const char * file;
		const char * function;
		unsigned long line;
		bool isArray;
	};

	typedef std::vector<ALLOC_INFO*> AllocList;

	AllocList *allocList;

	char * allocDeleteFile;
	char * allocDeleteFunction;
	int allocDeleteLine = 0;

	bool leakDebugFileOpened = false;

	// outputs string to log file
	void OutputDebugString(char *debugStr)
	{
		// clear the output file if it hasn't been opened yet
		if(!leakDebugFileOpened)
		{
			FILE *out;
			if((out = fopen(CLEAN_MEM_LOG_FILE, "w")) != NULL)
				fclose(out);
			leakDebugFileOpened = true;
		}

		FILE *out;

		if((out = fopen(CLEAN_MEM_LOG_FILE, "a")) == NULL)
			fprintf(stderr, "Cannot open output file.\n");

		fputs(debugStr, out);

		fflush(out);
		fclose(out);
	};

	bool checkLeakScope = false;

	void BeginCheckingUnfreed(void)
	{
		checkLeakScope = true;
	}

	static void strcatAfter(char * str, const char * cat, int minFirstSegLen)
	{
		int len = strlen(str);

		if(len < minFirstSegLen)
		{
			for(int i = len ; i < minFirstSegLen ; i++)
				str[i] = ' ';

			str[minFirstSegLen] = '\0';
		}

		strcat(str, cat);
	}

	// shows a summary of unfreed memory
	void DumpUnfreed(void)
	{
		// don't dump anything if there were no memory leaks AND there is no existing dump file
		if(!allocList || allocList->empty())
		{
			FILE *test;
			if((test = fopen(CLEAN_MEM_LOG_FILE, "rb")) != NULL)
				fclose(test);
			else return;
		}

		AllocList::iterator i;
		unsigned long totalSize = 0;
		char buf[1024];

		sprintf(buf, "\n-----------------------------------------------------------\n");			OutputDebugString(buf);
		sprintf(buf, " Begin Dump of Unfreed Memory:     (Output ID = %d)\n", (time(NULL)/4)%8092);	OutputDebugString(buf);
		sprintf(buf, "-----------------------------------------------------------\n\n");			OutputDebugString(buf);

		if(!allocList || allocList->empty())
		{
			sprintf(buf, "No memory leaks detected!\n");
			OutputDebugString(buf);
		}
		else
		{
			int maxLineNumLength = 0;
			int maxFileLength = 0;
			int maxFunctionLength = 0;

			// calculate maximum filename and function name lengths for a pretty printout
			for(unsigned int i = 0 ; i < allocList->size() ; i++)
			{
				ALLOC_INFO * info = (*allocList)[i];
				char fileStr [128];
				strncpy(fileStr, info->file, 128);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				int fileLength = strlen(fileStrPtr);
				if(maxFileLength < fileLength)
					maxFileLength = fileLength;

				int functionLength = strlen(info->function);
				if(maxFunctionLength < functionLength)
					maxFunctionLength = functionLength;

				char numStr [64];
				sprintf(numStr, "%d", info->line);

				int lineNumLength = strlen(numStr);
				if(maxLineNumLength < lineNumLength)
					maxLineNumLength = lineNumLength;
			}

			// loop through and print all relevant info about each memory leak in chronological order:
			for(unsigned int i = 0 ; i < allocList->size() ; i++)
			{
				ALLOC_INFO * info = (*allocList)[i];
				char fileStr [128];
				strncpy(fileStr, info->file, 128);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				bool alreadyDone = false;
				int bytes = 0;

				// for each allocList entry remaining (i.e. for each memory leak),
				for(unsigned int j = 0 ; j < allocList->size() ; j++)
				{
					ALLOC_INFO * info2 = (*allocList)[j];
					if(info->line == info2->line
					&& stricmp(info->file, info2->file) == 0) // if they're the same entries,
					{
						if(j < i)
						{
							alreadyDone = true; // already included this entry in a multi-entry output
							break;
						}
						else
						{
							bytes += info2->size; // increment number of entries this output corresponds with
						}
					}
				}

				if(alreadyDone)
					continue; // already included this entry in a multi-entry output

				char tempStr [256];

				sprintf(tempStr, "line %d ", info->line);
				strcpy(buf, tempStr);

				sprintf(tempStr, "of %s ", fileStrPtr);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1);

				sprintf(tempStr, "in %s ", info->function);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1);

				sprintf(tempStr, "%d bytes\n", bytes);
				strcatAfter(buf, tempStr, 5+maxLineNumLength+1 + 3+maxFileLength+1 + 3+maxFunctionLength+1);

				OutputDebugString(buf);
				totalSize += bytes;
			}
		}

		sprintf(buf, "\nTotal Unfreed: %d bytes\n\n", totalSize);						OutputDebugString(buf);
		sprintf(buf, "-----------------------------------------------------------\n");	OutputDebugString(buf);
		sprintf(buf, " End Dump of Unfreed Memory\n");									OutputDebugString(buf);
		sprintf(buf, "-----------------------------------------------------------\n");	OutputDebugString(buf);

		checkLeakScope = false;
	};

	// adds an allocated memory entry
	// TODO: if memory has already been allocated on the same line of the same file,
	//       this should just increase the 'size' value of the existing track
	//       as opposed to adding a whole new track and summing up their sizes later which is done now
	void AddTrack(unsigned long addr,  unsigned long asize,  const char *fname, const char *funcName, unsigned long lnum, bool array)
	{
		ALLOC_INFO *info;

		if(!allocList) {
			allocList = new(AllocList);
		}

		info = new(ALLOC_INFO);
		info->address = addr;
		info->file = fname;
		info->function = funcName;
		info->line = lnum;
		info->size = asize;
		info->isArray = array;
		allocList->push_back(info);

#ifdef DEBUG_OUTPUT_ALL_ALLOCS
		totalAllocated += asize;
#	ifdef DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
		if(GetAsyncKeyState(VK_CONTROL) != 0)
#	endif
		{
			char tempStr [256], buf [1024];

			sprintf(tempStr, "ALLOC line %d ", info->line);
			strcpy(buf, tempStr);

			sprintf(tempStr, "of %s ", info->file);
			strcat(buf, tempStr);

			sprintf(tempStr, "in %s ", info->function);
			strcat(buf, tempStr);

			sprintf(tempStr, "%d bytes / %d total\n", info->size, totalAllocated);
			strcat(buf, tempStr);

			OutputDebugString(buf);
		}
#		endif
	};

	// removes an allocated memory entry (when mem is de-allocated)
	void RemoveTrack(unsigned long addr, bool array)
	{
		AllocList::iterator i;

		if(!allocList)
			return;

		for(i = allocList->begin(); i != allocList->end(); i++)
		{
			ALLOC_INFO *info = *i;

 			if(info->address == addr)
			{
				// if this assert fails, we mismatched new[] with delete, or new with delete[]
				assert(info->isArray == array);

		#ifdef DEBUG_OUTPUT_ALL_ALLOCS
				totalAllocated -= info->size;
		#	ifdef DEBUG_OUTPUT_ALL_ALLOCS_ONLY_ON_CTRL
				if(GetAsyncKeyState(VK_CONTROL) != 0)
		#	endif
				{
					char tempStr [256], buf [1024];

					sprintf(tempStr, "FREE at line %d ", allocDeleteLine);
					strcpy(buf, tempStr);

					sprintf(tempStr, "of %s ", allocDeleteFile);
					strcat(buf, tempStr);

					sprintf(tempStr, "in %s ", allocDeleteFunction);
					strcat(buf, tempStr);

					sprintf(tempStr, "-%d bytes / %d total\n", info->size, totalAllocated);
					strcat(buf, tempStr);

					OutputDebugString(buf);
				}
		#		endif

				allocList->erase(i);
				return;
			}
		}

/*
		if(checkLeakScope)
		{
			char buf [1024];
			sprintf(buf, "Invalid Removal: address %d, via delete%s\n", addr, array?"[]":"");
			OutputDebugString(buf);
		}*/
	};

	void * __malloc(size_t size, const char *file, int line, const char *function)
	{
		void * ptr;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

		ptr = (void *)malloc(size);

		if(checkLeakScope)
			AddTrack((unsigned long)ptr, size, file, function, line, false);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, file, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "NEW of %d bytes FAILED on line %d of %s\tin %s\n", size, line, fileStrPtr, function);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
		}
		#endif

		return ptr;
	}

	void __free(void * mem)
	{
		if(!mem)
			return;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
			// If it breaks here, hit "Run" again then check the top of leaks.txt for more info
			// NOTE: a "user breakpoint" hit may be caused by forgetting to compile with Runtime = "Multi-threaded Debug DLL (/MDd)"
			if(checkLeakScope && _CrtIsValidHeapPointer(mem) == false)
				throw "Invalid Pointer";

		#endif

		if(checkLeakScope)
			RemoveTrack((unsigned long)mem, false);

		free(mem);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, allocDeleteFile, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "DELETE of 0x%x FAILED on line %d of %s\tin %s\n", (unsigned long)mem, allocDeleteLine, fileStrPtr, allocDeleteFunction);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and maybe the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
		}
		#endif
	}

	// overloaded new
	void * operator new(size_t size, const char *file, int line, const char *function)
	{
		return __malloc(size,file,line,function);
	};

	// overloaded new[]
	void * operator new[](size_t size, const char *file, int line, const char *function)
	{
		void * ptr;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
		#endif

		ptr = (void *)malloc(size);

		if(checkLeakScope)
			AddTrack((unsigned long)ptr, size, file, function, line, true);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, file, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "NEW[] of %d bytes FAILED on line %d of %s\tin %s\n", size, line, fileStrPtr, function);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and maybe the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
		}
		#endif

		return ptr;
	};

	// overloaded delete
	void operator delete(void * mem)
	{
		__free(mem);
	};

	// overloaded delete
	void operator delete[](void * mem)
	{
		if(!mem)
			return;

		#ifndef CLEAN_MEM_NO_VERIFY
		try
		{
			// If it breaks here, hit "Run" again then check the top of leaks.txt for more info
			// NOTE: a "user breakpoint" hit may be caused by forgetting to compile with Runtime = "Multi-threaded Debug DLL (/MDd)"
			if(checkLeakScope && _CrtIsValidHeapPointer(mem) == false)
				throw "Invalid Pointer";

		#endif


		if(checkLeakScope)
			RemoveTrack((unsigned long)mem, true);


		free(mem);

		#ifndef CLEAN_MEM_NO_VERIFY
		}
		catch(...)
		{
			if(checkLeakScope && allocDeleteLine != 0)
			{
				char fileStr [128];
				strncpy(fileStr, allocDeleteFile, 127);
				char * fileStrPtr = (char*) strrchr(fileStr, (int)'\\');
				if(fileStrPtr) fileStrPtr++;
				else		   fileStrPtr = (char*) fileStr;

				char tempStr[2048];
				sprintf(tempStr, "DELETE[] of 0x%x FAILED on line %d of %s\tin %s\n", (unsigned long)mem, allocDeleteLine, fileStrPtr, allocDeleteFunction);
				OutputDebugString(tempStr);
///				DumpUnfreed();
///				#ifdef _DEBUG
///					DebugBreak(); // if it breaks here, check leaks.log and the rest of the call stack for more info
///					exit(-1);
///				#endif
				allocDeleteLine = 0;
			}
///			Engine.ErrorDialog(tempStr);
		}
		#endif

	};

#endif // !CLEAN_MEM_DISABLE

