//
// GameX - Initialization- and Shutdown-related WindowsDX Class Code
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 
// this is a version of GameX that has been modified (by Justin Pease)


// KNOWN ISSUES:

// R5G5B5 display mode (sometimes called 15-bit mode) may not be supported,
// because I've never actually seen a monitor that uses it.
// i.e., R5G5B5 mode *should* work but has not been tested, so it should be tested sometime soon.


#include "gamex-win-dx.hpp"

#include "gamex-win-dx-timer.inl"


CameraX * defaultCam = NULL;

// WindowsDX Constructor
WindowsDX::WindowsDX (void)
{	
	BeginCheckingUnfreed(); // start checking for memory leak / corruption as soon as possible
	atexit(DumpUnfreed); // stop checking for memory leak / corruption as late as possible

	win_request_xsize = -1;
	win_request_ysize = -1;
	win_status = GAMEX_NO_INIT;
	win_init_done = false;
	num_sounds = 0;
	max_sounds = 0;
///	sound_list = NULL;
	sound_support = false;
	drawLocked = false;
	win_warning_on_draw = false;
	strcpy(win_name, "GameX");
	draw_state = new GameXDrawState();
	draw_state_head = NULL;
	draw_state_tail = NULL;
	defaultCam = new CameraX();
	draw_state->cam = defaultCam;
	draw_state->dst = VIEWPORT;
	ResetDrawStates();
	win_err_debug = true;
#ifdef _DEBUG
	win_log_debug = true;
#else
	win_log_debug = false;
#endif
	for(int i=0 ; i<GAMEX_TOTAL_STATUS_NUMBER ; i++)
		win_status_problem_ignore[i] = false;
	win_appIconIDStr = IDI_APPLICATION;

	text_printer_img = NULL;
}

WindowsDX::~WindowsDX (void)
{
	SAFE_DELETE(draw_state);
	SAFE_DELETE(defaultCam);

	SAFE_DELETE(text_printer_img);
	// WindowsDX::CleanUp() does the main uninitializing
}


// This function must be called by the game's GameInit() function
// Parameters: Name, Flags, ViewWidth, ViewHeight, [optional:]	Speed, Smoothness, SkipLimit 
bool WindowsDX::Initialize (char * name, GameXInitFlags options, int xres, int yres, int speed)
{
	strcpy(win_name,name); // remember game name

	ResetRandomSeed(); // set the random number seed in case the game forgets to

	if(win_err_debug) {
		if(options & VIDEO_WINDOWED && options & VIDEO_FULLSCREEN) {
			MessageBox (win_hwnd, "WARNING: VIDEO_WINDOWED and VIDEO_FULLSCREEN are mutually exclusive flags.\nLeave out both flags if you want the user to choose.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
			options ^= VIDEO_WINDOWED;
			options ^= VIDEO_FULLSCREEN;
		}
	}

	quit_dialog = false;

	if(options & VIDEO_WINDOWED)
		win_request_full = 0;
	else if(options & VIDEO_FULLSCREEN)
		win_request_full = 1;
	else
		win_request_full = 2;

	if(options & VIDEO_NORESIZE) {
		win_request_can_resize = false;
	} else {
		win_request_can_resize = true;
	}

	if(options & VIDEO_LOWRES) {
		win_request_low_resolution = true;
	} else {
		win_request_low_resolution = false;
	}


	if(win_request_full == 2) {
		int style = MB_YESNOCANCEL|MB_ICONQUESTION;

		if(win_log_debug)
			style |= MB_DEFBUTTON2;

		int choice = MessageBox (NULL, "Do you want to play this game in fullscreen mode?", win_name, style) ;
		if(choice == IDYES)
			win_request_full = 1;
		else if(choice == IDNO)
			win_request_full = 0;
		else if(choice == IDCANCEL)
			Quit(true);
	}

	if(options & VIDEO_ALLOWREFRESHSYNC)
		win_request_vsync = 1;
	else
		win_request_vsync = 0;

	if(options & VIDEO_SOFTWAREONLY) {
		MessageBox (win_hwnd, "WARNING: Entering software-only mode because VIDEO_SOFTWAREONLY was turned on.\nThis is just a reminder: don't forget to remove that flag.", "GameX Notice",MB_OK|MB_ICONINFORMATION) ;
		win_request_software_only = true;
	}
	else
		win_request_software_only = false;

	if(options & VIDEO_SLOWDEBUGDRAWING) {
		MessageBox (win_hwnd, "WARNING: Entering ultra-slow drawing mode because the VIDEO_SLOWDEBUGDRAWING was set.\nJust a reminder: don't forget to remove that flag.", "GameX Notice",MB_OK|MB_ICONINFORMATION) ;
		win_request_slow_debug_drawing = true;
//		speed = 5;
	}
	else
		win_request_slow_debug_drawing = false;

	win_request_delete_mdata = true;

	win_request_bpp = 0;
	if(options & VIDEO_16BIT) {
		win_request_bpp = 16; // 16 bit color is fully supported by GameX
	} else if(options & VIDEO_24BIT) {
		win_request_bpp = 24;
		MessageBox (win_hwnd, "WARNING: 24-bit RGB color mode has not yet been tested.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
	} else if(options & VIDEO_32BIT) {
		win_request_bpp = 32; // 32 bit color is fully supported by GameX -- may be a little slow though
	}

	if(win_err_debug) {
		if(1 < /*!!(options & VIDEO_8BIT)+*/!!(options & VIDEO_16BIT)+!!(options & VIDEO_24BIT)+!!(options & VIDEO_32BIT)) {
			MessageBox (win_hwnd, "ERROR: You cannot select more than one video mode at once.\nPlease initialize GameX with only one VIDEO_??BIT flag.", "GameX Error",MB_OK|MB_ICONSTOP) ;
		} else if(win_request_bpp == -1) {
			MessageBox (win_hwnd, "WARNING: You should select a video mode by initializing GameX with a video color depth flag.\nFor example,\nGameX.Initialize(\"My Game\",VIDEO_16BIT,640,480);\nor\nGameX.Initialize(\"My Game\",VIDEO_32BIT|VIDEO_FULLSCREEN,800,600);", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		}
	}

	win_request_xsize = xres;
	win_request_ysize = yres;

	win_cps = speed; // see WinGameRun() for usage

	if(win_cps != 0) // if it isn't set to unmanaged
	if(win_cps < 5 || win_cps > 600) {
		if(win_err_debug) {
			char disp [500] ;
			sprintf (disp, "WARNING: You have specified %d CPS, which is invalid. The CPS must be at least 5, at most 600, and normally should be between 30 and 120.", win_cps) ;
			MessageBox (win_hwnd, disp, "GameX Warning",MB_OK|MB_ICONWARNING) ;
		}
		if(win_cps < 5) win_cps = 5;
		else if(win_cps > 600) win_cps = 600;
	}

	win_fps = -1; // default of targeting the best possible framerate that's visible
	win_max_cpf = max(3,win_cps/5); // default to skip no more than 1/5 of a second of frames per instant


	if(options & AUDIO_DISABLE)
		win_request_no_sound = true;
	else
		win_request_no_sound = false;

	if(options & RUN_USEESCAPEKEY)
		win_request_use_escape = true;
	else
		win_request_use_escape = false;

	if(options & RUN_NOALTENTERTOGGLE)
		win_request_use_alt_enter = true;
	else
		win_request_use_alt_enter = false;

	if(options & RUN_AUTOSHOWINFO)
		win_request_show_run_info = true, win_log_debug = true;
	else
		win_request_show_run_info = false;

	if(options & RUN_BACKGROUND)
		win_request_no_auto_pause = true;
	else
		win_request_no_auto_pause = false;

	if(options & RUN_ALWAYSKEEPCURSOR && options & RUN_ALWAYSHIDECURSOR) {
		if(win_err_debug)
			MessageBox (win_hwnd, "WARNING: RUN_ALWAYSKEEPCURSOR and RUN_ALWAYSHIDECURSOR are mutually exclusive flags.", "GameX Warning",MB_OK|MB_ICONWARNING) ;
		options ^= RUN_ALWAYSKEEPCURSOR;
		options ^= RUN_ALWAYSHIDECURSOR;
	}

	if(options & RUN_ALWAYSKEEPCURSOR)
		win_request_no_hide_cursor = true;
	else
		win_request_no_hide_cursor = false;

	if(options & RUN_ALWAYSHIDECURSOR)
		win_request_no_show_cursor = true;
	else
		win_request_no_show_cursor = false;

	if(options & RUN_NOMOUSEINPUT)
		win_request_no_mouse_input = true;
	else
		win_request_no_mouse_input = false;

	if(options & RUN_NOCONFIRMQUIT)
		win_request_no_confirm_quit = true;
	else
		win_request_no_confirm_quit = false;

	bool success = true;

//	debug.Open ("debug.txt") ;
	if(win_log_debug)
		debug.Output ("WindowsDX: Initialization started.\n") ; 

	if (InitWindowSetup()==false) success = false;
	if(success)
	if (InitDirectX()==false) success = false;
	
	if(!success) {
		ReportProblem(GetStatus());
		debug.Output ("WindowsDX: Initialization had problems.\n") ;
		return false;
	}

	DoInitTests();

	DisplayLoadingWithValue (78);

	if(!tp.IsInitialized())
	{
		FILE * file = NULL;
		char str [256];
		if(!file) strcpy(str, "text.gif"), file = fopen(str,"rb");
		if(!file) strcpy(str, "text.jpg"), file = fopen(str,"rb");
		if(!file) strcpy(str, "text.bmp"), file = fopen(str,"rb");
		if(!file) strcpy(str, "text.png"), file = fopen(str,"rb");
		if(file)
		{
			fclose(file);
			text_printer_img = new ImageX();
			text_printer_img->Load(str,true);
			tp.Initialize(text_printer_img, 16);
		}
	}

	DisplayLoadingWithValue (80);

	if(win_log_debug)
		debug.Output ("WindowsDX: Initialization finished OK.\n") ;

	win_init_done = true;

	return true;
}

void WindowsDX::DoInitTests (void)
{
	win_status = GAMEX_CREATING_TEST_SURFACE;

	// test of blurring speed:
	win_blur_supported = true; // temporarily true to allow testing
	int trials = 10;
	hyper timer = GetClock();
	for(int i = 0 ; i < trials ; i++) {
		BlurScreen(0.51f);
	}
	timer = GetClock() - timer;
	double testTime = (double)(timer) / (double)GetClocksPerSec() / (double)trials;
	double minTime = 0.025*(double)(win_width*win_height)/(double)(800*600); // minimum acceptable time for test (5x slower than decent speed of 0.005 seconds/trial/(800*600)pixels)
	if(testTime > minTime) {
		debug.Output("WinDX: Blur speed test FAILED, blurring disabled.\n");
		win_blur_supported = false;
	} else {
		win_blur_supported = true;
	}



	// fill the background with 50% gray

	// (make sure it worked, while we're at it:)
	int r=0,g=0,b=0;
	int tol = 24; // tolerance allowed in each direction from expected color component -- necessary because of color format imprecision as well as slight deviations in video card processing results

	// create test images:
	ImageX testImage, testImage2, testAlphaImage, testDDImage;
	testImage.Create(16,16);
	testImage.Fill(255,0,0); // fill with bright red
	testImage2.Create(16,16);
	testImage2.Fill(192,128,64); // fill with brownish color
	testAlphaImage.Create(16,16,true);
	testAlphaImage.Fill(255,0,0,255); // fill with bright red (full opacity)
	testDDImage.Create(64,64);
	testDDImage.ConvertUsageTo(LOAD_NON3D);
	testDDImage.Fill(255,0,0); // fill with bright red

	// Draw Mode Tests:
	// Some video cards handle certain drawing modes improperly,
	// so we test for correctness here and enable work-arounds as necessary:

	// Subtractive (DRAW_SUBTRACT mode) Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	win_subtract_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_SUBTRACT);
	DrawImage(&testImage, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (0,128,128) after this line
	EndPixelAccess();
	if(r>0+tol || r<0-tol || g>128+tol || g<128-tol || b>128+tol || b<128-tol) {
		if(win_fullscreen) ToggleFullscreen();
		
		FillScreen(ColorX(128,128,128));
		SetDrawMode(DRAW_SUBTRACT);
		DrawImage(&testAlphaImage, 0,0);

		AccessPixels(VIEWPORT);
		ReadPixel(0,0,r,g,b); // (r,g,b) should equal (0,128,128) after this line
		EndPixelAccess();

		if(r>0+tol || r<0-tol || g>128+tol || g<128-tol || b>128+tol || b<128-tol) {
			debug.Output("WinDX: Subtractive drawing test FAILED, black-alpha work-around enabled.\n");
			win_subtract_workaround = 2; // black-alpha subtraction simulation required (worst case)  uses image with all black and varying alpha image to darken destination non-subtractively
		} else {
			debug.Output("WinDX: Subtractive drawing test FAILED, 32-bit work-around enabled.\n");
			win_subtract_workaround = 1; // subtractive drawing works with alpha images although not normally, so enable workaround that converts images to alpha if drawn subtractively
		}
		
	} else {
		win_subtract_workaround = 0; // no workaround required
	}

	// Multiplicatively Additive (DRAW_INTENSIFY mode) Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	win_intensify_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_INTENSIFY);
	DrawImage(&testImage, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (255,128,128) after this line
	EndPixelAccess();
	if(r>255+tol || r<255-tol || g>128+tol || g<128-tol || b>128+tol || b<128-tol) {
		debug.Output("WinDX: Intensify drawing test FAILED, additive work-around enabled.\n");
		win_intensify_workaround = 1; // workaround: simply use regular addition instead of multiplicative addition
	} else {
		win_intensify_workaround = 0; // no workaround required
	}

	// Color-valued Alpha (DRAW_GHOST mode) Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	win_ghost_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_GHOST);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (192,128,112) after this line
	EndPixelAccess();
	if(r>192+tol || r<192-tol || g>128+tol || g<128-tol || b>112+tol || b<112-tol) {
		debug.Output("WinDX: Ghost drawing test FAILED, alpha work-around enabled.\n");
		win_ghost_workaround = 1; // workaround required
	} else {
		win_ghost_workaround = 0; // no workaround required
	}

	// DRAW_BURN Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	win_burn_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_BURN);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (64,128,144) after this line
	EndPixelAccess();
	if(r>64+tol || r<64-tol || g>128+tol || g<128-tol || b>144+tol || b<144-tol) {
		debug.Output("WinDX: Burn drawing test FAILED, invert+subtract work-around enabled.\n");
		win_burn_workaround = 1; // workaround required
	} else {
		win_burn_workaround = 0; // no workaround required
	}

	
	// DRAW_ADDSOFT Drawing Test:

	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	win_addsoft_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_ADDSOFT);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (224,192,160) after this line
	EndPixelAccess();
	if(r>224+tol || r<224-tol || g>192+tol || g<192-tol || b>160+tol || b<160-tol) {
		debug.Output("WinDX: Soft additive drawing test FAILED, regular addition work-around enabled.\n");
		win_addsoft_workaround = 1; // workaround: simply use regular addition instead
	} else {
		win_addsoft_workaround = 0; // no workaround required
	}


	// DRAW_ADDSHARP Drawing Test:

	FillScreen(ColorX(0,0,0)); // (on top of a black background)

	win_addsharp_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_ADDSHARP);
	DrawImage(&testImage2, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should equal (255,255,64) after this line
	EndPixelAccess();
	if(r>255+tol || r<255-tol || g>255+tol || g<255-tol || b>64+tol || b<64-tol) {
		debug.Output("WinDX: Sharp additive drawing test FAILED, regular addition work-around enabled.\n");
		win_addsharp_workaround = 1; // workaround: simply use regular addition instead
	} else {
		win_addsharp_workaround = 0; // no workaround required
	}


	// Alpha-blending Test:
	// (some video cards don't apply alpha blending properly except on alpha images or non-textured polygons)
	FillScreen(ColorX(128,128,128)); // (on top of a 50% gray background)

	win_alpha_blend_workaround = 0; // try without a workaround first
	SetDrawMode(DRAW_PLAIN);
	SetDrawShading(ColorX(255,255,255,128));
	DrawImage(&testImage, 0,0);

	AccessPixels(VIEWPORT);
	ReadPixel(0,0,r,g,b); // (r,g,b) should about equal (192,64,64) after this line
	EndPixelAccess();
	if(r>192+tol || r<192-tol || g>64+tol || g<64-tol || b>64+tol || b<64-tol) {
		FillScreen(ColorX(128,128,128));
		SetDrawMode(DRAW_PLAIN);
		SetDrawShading(ColorX(255,255,255,128));
		DrawImage(&testAlphaImage, 0,0);

		AccessPixels(VIEWPORT);
		ReadPixel(0,0,r,g,b); // (r,g,b) should about equal (192,64,64) after this line
		EndPixelAccess();

		if(r>192+tol || r<192-tol || g>64+tol || g<64-tol || b>64+tol || b<64-tol) {
			debug.Output("WinDX: Alpha-blend shading test FAILED, some alpha shading disabled.\n");
			win_alpha_blend_workaround = 0; // no workaround known in this case
		} else {
			debug.Output("WinDX: Alpha-blend shading test FAILED, 32-bit work-around enabled.\n");
			win_alpha_blend_workaround = 1; // if works on alpha images, convert images to 32-bit if drawn with alpha shading
		}
	} else {
		win_alpha_blend_workaround = 0; // no workaround required
	}

	double nonStretchTime, stretchTime;
	{
		int trials = 50;
		for(int i = 0 ; i < trials ; i++)
			DrawTexturedRect(&testDDImage, 0,0,64,64);
	}
	{
		// test of non-stretching speed:
		int trials = 25;
		hyper timer = GetClock();
		for(int i = 0 ; i < trials ; i++) {
			DrawTexturedRect(&testDDImage, 0,0,64,64);
		} 
		timer = GetClock() - timer;
		nonStretchTime = 1000.0 * (double)(timer) / (double)GetClocksPerSec() / (double)trials;
	}
	{
		// test of stretching speed:
		int swidth = min(400,GameX.GetWidth());
		int sheight = min(400,GameX.GetHeight());
		int trials = 25;
		hyper timer = GetClock();
		for(int i = 0 ; i < trials ; i++) {
			DrawTexturedRect(&testDDImage, 0,0,swidth,sheight);
		}
		timer = GetClock() - timer;
		stretchTime = 1000.0 * (double)(timer) / (double)GetClocksPerSec() / (double)trials;
	}

	if(stretchTime > nonStretchTime*10.0f) {
		win_resize_workaround = 1;

		RECT rect; // in case window began as too large already:
		GetClientRect (win_hwnd, &rect) ;
		win_xoffset = (rect.right-rect.left - win_width) / 2;
		win_yoffset = (rect.bottom-rect.top - win_height) / 2;				
		win_realx = win_width;
		win_realy = win_height;
	}

	ClearScreen();

	win_status = GAMEX_READY;
}

bool WindowsDX::InitWindowSetup (void)
{
	if (InitWindowClass()==false)		return false;
	if (InitWindows()==false)			return false;
	if (WindowDraw()==false)			return false;
//	SendMessage(win_hwnd,WM_SETICON,ICON_SMALL,-1); // remove icon since I can't seem to get it to display the one it should
	return true;
}

bool WindowsDX::InitDirectX (void)
{	
	if(FAILED(CoInitialize(NULL))) {
		win_status = GAMEX_DDSTART_FAILED;
		return false;
	}

	bool success = true;

	if(GetAsyncKeyState(VK_ESCAPE) & 0x8000)
		Quit(true);

	if(InitDirectDraw()==false) {
		success = false;
		ReportProblem();
		if(win_err_debug)
			debug.Output("WinDX: DirectDraw init failed.\n");
	} else {
		if(win_log_debug)
			debug.Output("WinDX: DirectDraw init succeeded.\n");
	}

	DisplayLoadingWithValue (10);

	if(InitDirect3D()==false) {
		success = false;
		ReportProblem();
		if(win_err_debug)
			debug.Output("WinDX: Direct3D init failed.\n");
	} else {
		if(win_log_debug)
			debug.Output("WinDX: Direct3D init succeeded.\n");
	}

	DisplayLoadingWithValue (20);

	if(InitDirectInput()==false) {
		success = false;
		ReportProblem();
		if(win_err_debug)
			debug.Output("WinDX: DirectInput init failed.\n");
	} else {
		if(win_log_debug)
			debug.Output("WinDX: DirectInput init succeeded.\n");
	}

	DisplayLoadingWithValue (30);
	
	if(InitDirectSound()==false) {
		success = false;
		ReportProblem();
		if(win_err_debug)
			debug.Output("WinDX: DirectSound init failed.\n");
	} else {
		if(win_log_debug)
			debug.Output("WinDX: DirectSound init succeeded.\n");
	}
	
	DisplayLoadingWithValue (40);

	if(InitDirectMusic()==false) {
		success = false;
		ReportProblem();
		if(win_err_debug)
			debug.Output("WinDX: DirectMusic init failed.\n");
	} else {
		if(win_log_debug)
			debug.Output("WinDX: DirectMusic init succeeded.\n");
	}

	DisplayLoadingWithValue (50);

	if(TestDirectShow()==false) {
		mp3_support = false;
		success = false;
		ReportProblem();
		if(win_err_debug)
			debug.Output("WinDX: DirectShow init failed.\n");
	} else {
		mp3_support = true;
		if(win_log_debug)
			debug.Output("WinDX: DirectShow init succeeded.\n");
	}

	DisplayLoadingWithValue (60);

	InitVSync(); // should be after other initializations

	DisplayLoadingWithValue (70);

	return success;
}

void WindowsDX::InitWindowsInstance (HINSTANCE hThisInst, int nWinMode)
{
	win_hinst = hThisInst;
	win_hmode = nWinMode;
}

extern LRESULT CALLBACK WindowFunc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

bool WindowsDX::InitWindowClass (void)
{
	WNDCLASS wcl;
	wcl.hCursor 		= LoadCursor(NULL, IDC_ARROW) ;
	wcl.hIcon			= LoadIcon(NULL, win_appIconIDStr);
	wcl.lpszMenuName	= "Menu" ;
	wcl.lpszClassName	= "GameX" ;
	wcl.hInstance		= win_hinst;
	wcl.style			= CS_VREDRAW | CS_HREDRAW;
	if(!win_request_no_mouse_input) wcl.style |= CS_DBLCLKS; // enable double-click recognition in the game window
	wcl.lpfnWndProc 	= (WNDPROC) WindowFunc;
	wcl.cbClsExtra		= 0;
	wcl.cbWndExtra		= 0;
	wcl.hbrBackground	= (HBRUSH) GetStockObject(BLACK_BRUSH) ;
	win_resize_workaround = 0;

	if (!RegisterClass(&wcl)) { 
		win_status = GAMEX_INITCLASS_FAILED;
		return false;
	}
	win_status = GAMEX_INITCLASS_OK;
	return true;
}

bool WindowsDX::InitWindows (void)
{
	DWORD style;

	// Initialize View Format (Bits-Per-Pixel)
	if(win_request_bpp != 0)
		win_screenbpp = win_request_bpp; // set bpp to requested setting
	else {
		HDC hdc = GetDC (win_hwnd) ; // current setting requested
		win_screenbpp = GetDeviceCaps (hdc, BITSPIXEL) ;
		ReleaseDC (win_hwnd, hdc) ;
	}

	// Initialize View Size & FullScreen Mode
	win_screenx = GetSystemMetrics(SM_CXSCREEN) ;
	win_screeny = GetSystemMetrics(SM_CYSCREEN) ;
	win_fullscreen = win_request_full ? true : false;
	win_no_acc_err = false;

	if(win_request_xsize == 0 && win_request_ysize == 0)
	{
		win_request_xsize = win_screenx;
		win_request_ysize = win_screeny;

		if(win_request_xsize / 3 > win_request_ysize * 4)
			win_request_xsize = win_request_ysize * 4 / 3;
		else
			win_request_ysize = win_request_xsize * 3 / 4;
	}
	else
	{
		if(win_request_xsize == 0) win_request_xsize = win_screenx;
		if(win_request_ysize == 0) win_request_ysize = win_screeny;
	}

	bool invalid_size = false;
	if(win_request_xsize > 2048) win_request_xsize = 2048, invalid_size=true;
	if(win_request_ysize >= 2048) win_request_ysize = 2047, invalid_size=true;
	if(win_request_xsize < 64) win_request_xsize = 64, invalid_size=true;
	if(win_request_ysize < 64) win_request_ysize = 64, invalid_size=true;

	if(win_err_debug) {
		if(invalid_size) {
			char * str = "Warning: Invalid game window size requested.\nIt must be between 64x64 and 2048x2047 pixels.";
			MessageBox (win_hwnd, str, "GameX Notice", MB_ICONWARNING) ;
			debug.Output("%s\n",str);
		}
	}

	// Initialize View Width & Height
	if (win_fullscreen==TRUE)	{
		style = WS_POPUP;	
		win_width = win_request_xsize;
		win_height = win_request_ysize;
	} else {
		style = WS_OVERLAPPED | WS_SYSMENU | WS_MINIMIZEBOX;
		if(win_request_can_resize) style |= WS_SIZEBOX | WS_MAXIMIZEBOX;
		win_width = win_request_xsize;		
		win_height = win_request_ysize;
	}

	win_realx = win_screenx;
	win_realy = win_screeny;
	int window_x = min(win_width, win_fullscreen ? win_screenx : win_realx-4); // temporary variable for actual width of window
	int window_y = min(win_height, win_fullscreen ? win_screeny : win_realy-32);

	// Initialize Drawing Viewport
	SetView(0,0,win_width,win_height);
		
	// Create window adjusting for desired client area size
	RECT	rect;
	int 	sx, sy;
	HWND	temp_hwnd;

	char title_str [256];
	sprintf(title_str, "%s - GameX 5.06", GetName());

	// Create temporary window to determine extended window size
	temp_hwnd = CreateWindowEx(WS_EX_APPWINDOW, "GameX", title_str, style, 0, 0, window_x, window_y, 0, 0, win_hinst, 0) ;
	if (temp_hwnd==NULL) {
		win_status = GAMEX_INITWINTEMP_FAILED;
		return false;		
	}
	win_status = GAMEX_INITWINTEMP_OK;
	GetClientRect (temp_hwnd, &rect) ;

	// set the rect to be the proper size and center it onscreen while we're at it
	sx = (window_x*2) - rect.right;
	sy = (window_y*2) - rect.bottom;
	if (win_fullscreen==TRUE) {
		rect.left = 0; rect.top = 0;
	} else {
		rect.left = (win_screenx/2) - (sx/2) ;
		rect.top = (win_screeny/2) - (sy/2) ;
	}	
	rect.right = rect.left + sx;		
	rect.bottom = rect.top + sy;

	// Create actual window (with correct size) which will be used

	win_hwnd = CreateWindowEx(
					WS_EX_APPWINDOW, 
					"GameX", 
					title_str, 
					style, 
					rect.left, rect.top, 
					rect.right-rect.left, rect.bottom-rect.top,
					0, 
					0, 
					win_hinst, 
					0) ;
	if (win_hwnd==NULL) {
		win_status = GAMEX_INITWINREAL_FAILED;		
		return false;
	}

	// Double check to make sure we have correct client rect size	
	GetClientRect (win_hwnd, &rect) ;	

	win_cursor_hidden = false;

	win_status = GAMEX_INITWINREAL_OK;
	return true;
}

bool WindowsDX::WindowDraw (void)
{
	ShowWindow (win_hwnd, win_hmode) ;
	if (UpdateWindow (win_hwnd)==false) {
		win_status = GAMEX_SHOWWIN_FAILED;		
		return false;
	}
	FlashWindow(win_hwnd,0);
	BringWindowToTop(win_hwnd);
	win_status = GAMEX_SHOWWIN_OK;
	return true;
}

HRESULT CALLBACK DisplayEnumCallback(LPDDSURFACEDESC2 lpDDModeDesc, LPVOID lpContext)
{
	if(GameX.SupportsDepth(lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount)
	&& lpDDModeDesc->ddpfPixelFormat.dwRGBBitCount == GameX.GetBPP()) { // if the bit depth matches
		if(lpDDModeDesc->dwWidth == GameX.GetWidth() && lpDDModeDesc->dwHeight == GameX.GetHeight()) { // if the resolution matches
			GameX.win_realx = GameX.GetWidth();
			GameX.win_realy = GameX.GetHeight();
			return DDENUMRET_CANCEL; // exact match, end enumeration with requested screen width and height
		}

		if((int)(lpDDModeDesc->dwWidth) >= GameX.GetWidth() && (int)(lpDDModeDesc->dwHeight) >= GameX.GetHeight() // if (this bigger than requested
		&& (int)(lpDDModeDesc->dwWidth) <= GameX.win_realx && (int)(lpDDModeDesc->dwHeight) <= GameX.win_realy // and this smaller than so far)
		|| (GameX.win_realx 			<  GameX.GetWidth() || GameX.win_realy				<  GameX.GetHeight()) // or (so far smaller than requested
		&& (int)(lpDDModeDesc->dwWidth) >= GameX.win_realx && (int)(lpDDModeDesc->dwHeight) >= GameX.win_realy) { // and this smaller than so far)
			GameX.win_realx = (int)(lpDDModeDesc->dwWidth);
			GameX.win_realy = (int)(lpDDModeDesc->dwHeight);
			return DDENUMRET_OK; // best match yet but not exact, remember this resolution and continue enumeration
		}
	}

	return DDENUMRET_OK; // rejected display mode, continue enumeration (or end automatically if there's no more to enumerate)
}

HRESULT CALLBACK DevicesEnumCallback(LPSTR lpDeviceDescription, LPSTR lpDeviceName, LPD3DDEVICEDESC7 lpD3DDeviceDesc, LPVOID lpContext)
{
	if (IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DNullDevice))
		return DDENUMRET_OK;

	if (!GameX.SimulatingSoftwareOnly() && IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DTnLHalDevice) )
	{
		GameX.d3d_device_desc = lpD3DDeviceDesc;
		return D3DENUMRET_CANCEL; // done
	}
	else if (!GameX.SimulatingSoftwareOnly() && IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DHALDevice) )
	{
		GameX.d3d_device_desc = lpD3DDeviceDesc;
		return D3DENUMRET_OK;
	}
	else if (IsEqualIID(lpD3DDeviceDesc->deviceGUID, IID_IDirect3DRGBDevice))
	{
		GameX.d3d_device_desc = lpD3DDeviceDesc;
		return D3DENUMRET_OK;
	}
	return D3DENUMRET_OK;
}

HRESULT CALLBACK ZBufferFormatEnumCallback (LPDDPIXELFORMAT lpDDPixFmt, LPVOID lpContext)
{
	GameX.d3d_zbuffer_format = *lpDDPixFmt;
	if(lpDDPixFmt != NULL && lpDDPixFmt->dwZBufferBitDepth == 16) return D3DENUMRET_CANCEL;

	return D3DENUMRET_OK;
}

bool WindowsDX::InitDirectDraw (void)
{
	dd_last_error = DD_OK;
	dd_was_lost = false;

	HRESULT status = DirectDrawCreateEx(NULL, (void **) &dd_main, IID_IDirectDraw7, NULL);

	if(FAILED(status)) {
		DirectDrawError(status);
		win_status = GAMEX_DDSTART_FAILED;
		return false;
	}
	win_status = GAMEX_DDSTART_OK;

	// set drawing state values to default
	draw_state->src_rect.left = 0;
	draw_state->src_rect.top = 0;
	draw_state->src_rect.right = DEFAULT;
	draw_state->src_rect.bottom = DEFAULT;
	draw_state->colors = ColorX();
	draw_state->color_effect = D3DTOP_MODULATE;
	draw_state->angle = 0.0f;
	draw_state->scalex = 1.0f;
	draw_state->scaley = 1.0f;
	draw_state->transx = 0.0f;
	draw_state->transy = 0.0f;
	draw_state->depth2D = 1.0f;
	draw_state->warp_mode = 0;
	draw_state->flags = DRAW_PLAIN;
	for(int i = 0 ; i < 4 ; i++)
		draw_state->warp[i] = 1.0f;
	draw_render_batch_open = false;
	draw_3D_depth_sort = 0;
	memory_used_current = 0;
	memory_used_max = 0;

	DDCAPS caps;
	ZeroMemory(&caps,sizeof(caps));
	caps.dwSize = sizeof(caps);
	if(SUCCEEDED(dd_main->GetCaps(&caps, NULL))) {

		if(win_log_debug) {
			char stringy [64];
			sprintf(stringy,"WinDX: Video memory available: %d megabytes.",caps.dwVidMemFree/1024/1024);
			debug.Output("%s\n",stringy);
		}

		int memfree = caps.dwVidMemFree;
		if(memfree < 1500000 && !win_request_software_only) { // if there are less than about 1.5 megabytes of VRAM available, assume there won't be much if any video acceleration
			char * str = "Warning: Lack of video acceleration detected!";
			MessageBox (win_hwnd, str, "GameX Notice", MB_ICONWARNING) ;
			debug.Output("%s\n",str);
	/// 	if(memfree == 0) win_request_software_only = true; // assume no hardware if there's no hardware memory at all
		} else {
			int memtotal = caps.dwVidMemTotal;
			if(memtotal < 22000000 && !win_request_software_only) { // if there are less than about 22 megabytes of VRAM total, assume the video card is probably too old (yes, the memory number only gives a very, very rough estimate of how good the card is)
			char * str = "Warning: Your video card appears to be outdated.\nConsider updating it if this game plays slowly or looks odd.";
			MessageBox (win_hwnd, str, "GameX Notice", MB_ICONWARNING) ;
			debug.Output("%s\n",str);
			}
		}
	}
 
	if(InitDirectDrawMonitor() == false)
		return false;

	if(InitDirectDrawSurfaces() == false)
		return false;

	return true;
}

bool WindowsDX::ToggleFullscreen(bool immediate)
{
	if(win_log_debug)
		debug.Output ("WindowsDX: Toggling full-screen mode...\n") ;

	// Request full screen mode that's the opposite of the current one
	win_request_full = !win_fullscreen;

	if(immediate)
	{
		bool ok = ReInitScreen();

///		SetWarningOnDraw(false); // leave off for the cycle
///		GameDraw();

		return ok;
	}
	else
		return true;
}

bool WindowsDX::ChangeGameSize(int xres, int yres)
{
	if(xres == win_width && yres == win_height)
		return true; // return if already set to size requested

	win_request_xsize = xres;
	win_request_ysize = yres;

	return ReInitScreen();
}

bool WindowsDX::ReInitScreen(void)
{
	// Set the status to something else than GAMEX_READY so UpdateView knows to temporarily do nothing
	win_status = GAMEX_DDSTART_OK;

	// Uninitialize the DirectX components that depend on the current window:

	// DirectInput
	if (di_mouse)		di_mouse->Unacquire(),		di_mouse->Release(),		di_mouse = NULL;
	if (di_keyboard)	di_keyboard->Unacquire(),	di_keyboard->Release(), 	di_keyboard = NULL;
	if (di_main)		di_main->Release(), 		di_main = NULL;

	// Direct3D
	if (d3d_device) 	d3d_device->Release(),		d3d_device = NULL;	
	if (d3d_main)		d3d_main->Release(),		d3d_main = NULL;	

	// DirectDraw
	if (dd_back_buf)	dd_back_buf->Release(), 	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release(),	dd_front_buf = NULL;	
	
	// Hide the window
	ShowWindow(win_hwnd,false);

	// Re-Initialize the window
	if(InitWindows() == false) return false;
	if(WindowDraw() == false) return false;

	// Re-Initialize the necessary DirectX components:
	if(InitDirectDrawMonitor() == false) return false;
	if(InitDirectDrawSurfaces() == false) return false;
	if(InitDirect3D() == false) return false;
	if(InitDirectInput() == false) return false;
	InitVSync();

	// Reset the GameX status to be ready and return true since we succeeded:
	win_status = GAMEX_READY;
	return true;
}

bool WindowsDX::InitDirectDrawMonitor (void)
{
	int style;
	if(win_fullscreen)
		style = DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT;
	else
		style = DDSCL_NORMAL;

	HRESULT status = dd_main->SetCooperativeLevel (win_hwnd, style);

	if(FAILED(status)) {
		if(win_fullscreen) {
			style ^= (DDSCL_ALLOWMODEX | DDSCL_ALLOWREBOOT); // try it without a couple flags
			status = dd_main->SetCooperativeLevel (win_hwnd, style);
			if(FAILED(status)) {
				win_fullscreen = false; // try it in windowed mode (maybe this computer doesn't allow fullscreen)
				style = DDSCL_NORMAL;
				status = dd_main->SetCooperativeLevel (win_hwnd, style);
			}
		}
		if(FAILED(status)) { // if we still failed
			win_status = GAMEX_DDSETLEVEL_FAILED;
			return false;
		}
	}
	win_status = GAMEX_DDSETLEVEL_OK;

	// now set the display mode:
	if (win_fullscreen==true) {

		win_realx = win_realy = 0;

		// enumerate available display modes and decide on one of them that best matches win_width, win_height, and win_screenbpp
		status = dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);

		DirectDrawError(status); // in case of error or to prevent compiler mishandling aliasing after callbacks

		if(!win_realx || !win_realy) { // if DisplayEnumCallback didn't set a screen size,
			
			// maybe the requested bit depth isn't supported, try for something higher
			while(win_screenbpp <= 32 && (!win_realx || !win_realy || !SupportsDepth(win_screenbpp))) {
				win_screenbpp += 8;
				status = dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);
				DirectDrawError(status);
			}
			// if that didn't work, try for something lower
			while(win_screenbpp > 0 && (!win_realx || !win_realy || !SupportsDepth(win_screenbpp))) {
				win_screenbpp -= 4;
				status = dd_main->EnumDisplayModes(0, NULL, NULL, (LPDDENUMMODESCALLBACK2) &DisplayEnumCallback);
				DirectDrawError(status);
			}
			if(!win_realx || !win_realy) {
				win_status = GAMEX_SETMODE_FAILED;
				return false;
			}
		}
		
		status = dd_main->SetDisplayMode (win_realx, win_realy, win_screenbpp,0,0);

		win_screenx = GetSystemMetrics(SM_CXSCREEN) ;
		win_screeny = GetSystemMetrics(SM_CYSCREEN) ;

		if(FAILED(status)) {
			DirectDrawError(status);
			win_status = GAMEX_SETMODE_FAILED;
			return false;
		}
		win_status = GAMEX_SETMODE_OK;

	} else {

		if(win_screenbpp != GetDeviceCaps (GetDC (win_hwnd), BITSPIXEL)) { // if we aren't in desired bit mode
			if(SupportsDepth(win_screenbpp))
				status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ; // change to that many bpp
			else
				status = DDERR_GENERIC;
			if(FAILED(status)) {
				// maybe the requested bit depth isn't supported, try for something higher
				while(win_screenbpp <= 32 && (FAILED(status) || !SupportsDepth(win_screenbpp))) {
					win_screenbpp += 8;
					status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ;
				}
				// if that didn't work, try for something lower
				while(win_screenbpp > 0 && (FAILED(status) || !SupportsDepth(win_screenbpp))) {
					win_screenbpp -= 4;
					status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ;
				}
				if(FAILED(status)) {
					// try something hard-coded that should definitely work, as a last resort
					win_screenx = 640;
					win_screeny = 480;
					win_screenbpp = 16;
					status = dd_main->SetDisplayMode (win_screenx, win_screeny, win_screenbpp,0,0) ;
					if(FAILED(status)) {
						win_status = GAMEX_SETMODE_FAILED;
						return false; // still failed... nothing else we can do in this case
					}
				}
			}
			UpdateWindow (win_hwnd); // redraw the window borders after switching or they'll all be black for a while
		}
		win_realx = min(win_width, win_realx-4); 
		win_realy = min(win_height, win_realy-32);

		win_status = GAMEX_SETMODE_OK;
	}

	DWORD monitor_frequency = 0;
	dd_main->GetMonitorFrequency(&monitor_frequency);
	if(monitor_frequency == 0) monitor_frequency = 60; // just in case
	win_screenrefresh = monitor_frequency;
	win_vsync_enabled = false; // will enable later if we can

	if(win_log_debug)
		debug.Output("WinDX: Chosen display mode: %dx%d %d-bit %dHz %sexclusive.\n", win_screenx, win_screeny, win_screenbpp, win_screenrefresh,win_fullscreen?"":"non");

	win_xpos = 0;
	win_ypos = 0;
	
	return true;
}

bool WindowsDX::InitDirectDrawSurfaces (void)
{
	HRESULT status;
	DDSURFACEDESC2 ddsd;	
	POINT pt;

	if (dd_back_buf)	dd_back_buf->Release(), 	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release(),	dd_front_buf = NULL;	

	ZeroMemory (&ddsd, sizeof (ddsd)) ;
	ddsd.dwSize = sizeof (ddsd) ;
 
	// Set front surface description
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	// Try to create primary surface 
	status = dd_main->CreateSurface (&ddsd, &dd_front_buf, NULL);
	if (FAILED(status)) {			
		DirectDrawError(status);
		win_status = GAMEX_FRONTSURFACE_FAILED; // No good, return error
		return false;
	}
	
	win_status = GAMEX_FRONTSURFACE_OK;
	// Set back surface description
	ddsd.dwSize = sizeof (ddsd) ;
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_3DDEVICE;
	if(win_request_software_only) ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
	ddsd.dwWidth = win_width;
	ddsd.dwHeight = win_height;

	// Try to create secondary surface 
	status = dd_main->CreateSurface (&ddsd, &dd_back_buf, NULL);
	if (FAILED(status)) {
		DirectDrawError(status);
		win_status = GAMEX_BACKSURFACE_FAILED; // No good, return error
		return false;
	}

	win_status = GAMEX_BACKSURFACE_OK;

	// Get Surface Description (pixel format, pitch, etc.)	
	ddsd.dwSize = sizeof(ddsd) ;
	ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_PITCH | DDSD_PIXELFORMAT;
	dd_front_buf->GetSurfaceDesc(&ddsd) ;
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {
		if(win_log_debug)
			debug.Output ("WinDX: Front buffer is in video memory.\n") ;
	} else {
		if(win_err_debug)
			debug.Output ("WinDX: Front buffer is NOT IN VIDEO MEMORY.\n") ;
	}
	dd_back_buf->GetSurfaceDesc(&ddsd) ;
	if (ddsd.ddsCaps.dwCaps & DDSCAPS_VIDEOMEMORY) {
		win_back_buf_in_vid_mem = true;
		if(win_log_debug)
			debug.Output ("WinDX: Back buffer is in video memory.\n") ;
	} else {
		win_back_buf_in_vid_mem = false;
		if(win_err_debug)
			debug.Output ("WinDX: Back buffer is NOT IN VIDEO MEMORY.\n") ;
	}

	// Initialize Drawing Functions, Pixel Masks, Pitch and Width 
	// based on Surface Description 

	win_pitch = ddsd.lPitch;
	win_adjustedpitch = win_pitch * 8 / win_screenbpp;

	win_maskr = ddsd.ddpfPixelFormat.dwRBitMask;
	win_maskg = ddsd.ddpfPixelFormat.dwGBitMask;		
	win_maskb = ddsd.ddpfPixelFormat.dwBBitMask;
	
	win_shiftr = GetShift (win_maskr) ; // 11 for 16 bit, 10 for 15 bit, 16 for 24 or 32 bit
	win_shiftg = GetShift (win_maskg) ; // 5 for 16 or 15 bit, 8 for 24 or 32 bit
	win_shiftb = GetShift (win_maskb) ; // 0

	win_maxr = GetMax (win_maskr) ;
	win_maxg = GetMax (win_maskg) ;
	win_maxb = GetMax (win_maskb) ;

	win_maskminr = 256/(win_maskr>>win_shiftr);
	win_maskming = 256/(win_maskg>>win_shiftg);
	win_maskminb = 256/(win_maskb>>win_shiftb);

	NullDrawPixelFuncs();

	// now set a clipper to the back surface so we don't have to clip bitmap blits ourselves
	{
		RECT boundary[1] = {0,0,win_width,win_height};
		LPDIRECTDRAWCLIPPER clipper;
		LPRGNDATA rgn_data;
		if (FAILED(dd_main->CreateClipper(0,&clipper,NULL))) {
			win_status = GAMEX_CLIPPER_FAILED;
			return false;
		}
		int rectSize = sizeof(RECT);
		int rgnHeaderSize = sizeof(RGNDATAHEADER);
		rgn_data = (LPRGNDATA)malloc(rgnHeaderSize+rectSize);
		memcpy(rgn_data->Buffer, boundary, rectSize);
		rgn_data->rdh.dwSize = rgnHeaderSize;
		rgn_data->rdh.nRgnSize = rectSize;
		rgn_data->rdh.nCount = 1;
		rgn_data->rdh.iType = RDH_RECTANGLES;
		rgn_data->rdh.rcBound = boundary[0];
		if (FAILED(clipper->SetClipList(rgn_data,0)) || FAILED(dd_back_buf->SetClipper(clipper))) {
			free(rgn_data);
			win_status = GAMEX_CLIPPER_FAILED;
			return false;
		}
		free(rgn_data);
	}

	LPDIRECTDRAWCLIPPER clipper = NULL; // hold the clipper
	dd_main->CreateClipper(0,&clipper,NULL);
	clipper->SetHWnd(0, win_hwnd);
	dd_front_buf->SetClipper(clipper);

	// Precompute: y-table look up 
	for (int y=0; y<=win_height; y++)
		win_ytable[y] = y*win_adjustedpitch;
		
	// Precompute: screen size (total number of bytes)
	win_size = win_pitch*win_height;
	
	// Precompute: client rectangle
	
	RECT destrect;

	GetClientRect(win_hwnd, &destrect) ;

	win_realx = destrect.right - destrect.left; // necessary on occasions where the OS gives us a different window size
	win_realy = destrect.bottom - destrect.top; // than we requested, such as when the requested width is very small or very large

	if(!IsResizeWorkaroundEnabled()) {
		SetProportional();
	} else {
		win_xoffset = (win_realx - win_width) / 2;
		win_yoffset = (win_realy - win_height) / 2; 			
		win_realx = win_width;
		win_realy = win_height;
	}

	pt.x = win_xpos;
	pt.y = win_ypos;
	ClientToScreen(win_hwnd, &pt) ;
	OffsetRect(&destrect, pt.x, pt.y) ;
	win_xpos = pt.x;
	win_ypos = pt.y;
	
	win_active = true;	
	win_status = GAMEX_READY;

	ClearScreen() ;
	UpdateView() ;

	return true;
}

void WindowsDX::InitVSync (void)
{
	char str [128];
	if(win_log_debug)
		strcpy(str,"WinDX: Refresh synchronization OFF.\n");

	if(win_request_vsync && win_fps > win_screenrefresh || win_fps == -1)
		win_fps = win_screenrefresh; // set the frame rate to not exceed the refresh rate

	if(win_request_vsync) {
		
		// test scanline speed:
		// sample 80 times at 10 milliseconds of delay each,

		DWORD scanline [80]; // array of consecutive scanline locations
		int diff [80]; // array of valid scanline location differences
		int n; // number of valid (positive) differences
		int i;

		hyper clocksPerMillisec = GetClocksPerSec()/1000;
		for(i = 0 ; i < 80 ; i++) {
			hyper sleeptimer = GetClock();
			while(GetClock() - sleeptimer < clocksPerMillisec) ;

			dd_main->GetScanLine(&scanline[i]);
		}

		n = 0;

		for(i = 0 ; i < 80 ; i++) {
			if(i && scanline[i]) {
				if(scanline[i] > scanline[i-1]) {
					diff[n] = scanline[i] - scanline[i-1];
					n++;
				}
			}
		}

		if(n < 3)
			win_request_vsync = 0;
		else {
			int total = 0;

			for(int i = 0 ; i < n ; i++) {
				total += diff[i];
			}

			win_scanline_speed = total/n;
			if(win_scanline_speed > 0) {
				if(win_screeny/win_scanline_speed*win_screenrefresh > CLOCKS_PER_SEC) {
					if(win_log_debug)
						sprintf(str,"WinDX: Refresh synchronization OFF; %d scanlines per millisecond is too high.\n", win_scanline_speed);
				} else {
					win_vsync_enabled = true;
					if(win_log_debug)
						sprintf(str,"WinDX: Refresh synchronization on, %d scanlines per millisecond.\n", win_scanline_speed);
				}
			}
		}
	}

	if(win_log_debug) {
		debug.Output("WinDX: Timer resolution: %.0f ticks per second.\n", (double)GetClocksPerSec());
		debug.Output(str);
	}
}

bool WindowsDX::InitDirect3D (void)
{
	// decide on a max texture size we'll allow for D3D surfaces
	// 1024 is a good limit; any higher wastes much speed and memory
	// note that it may also be further limited by hardware capability
	SetImageResolutionLimit (1024);

	// if DirectDraw is working, use it to intiialize Direct3D:
	if(dd_main==NULL || FAILED(dd_main->QueryInterface(IID_IDirect3D7, (LPVOID *)& d3d_main))) {
		debug.Output ("Direct3D: Device query failed.\n") ;
		win_status = GAMEX_D3DSTART_FAILED;
		return false; 
	}


	HRESULT status;

	// Enumerate available 3D rendering devices and decide on one:
	status = d3d_main->EnumDevices((LPD3DENUMDEVICESCALLBACK7) &DevicesEnumCallback, NULL);
	GUID guid = d3d_device_desc->deviceGUID;

	d3d_zbuffer_enabled = false;

	// make sure not to use a z-buffer unless we have access to video hardware
	if(guid != IID_IDirect3DRGBDevice) {

		DDPIXELFORMAT pixelzbuffer;
		d3d_main->EnumZBufferFormats(guid, (LPD3DENUMPIXELFORMATSCALLBACK) &ZBufferFormatEnumCallback, (VOID*)&pixelzbuffer );

		DDSURFACEDESC2 ddsdZ;
		ZeroMemory(&ddsdZ, sizeof(DDSURFACEDESC2));
		ddsdZ.dwSize = sizeof(DDSURFACEDESC2);
		ddsdZ.dwWidth = win_width;
		ddsdZ.dwHeight = win_height;
		ddsdZ.ddpfPixelFormat = d3d_zbuffer_format;
		ddsdZ.dwFlags = DDSD_HEIGHT | DDSD_WIDTH | DDSD_CAPS | DDSD_PIXELFORMAT;
		ddsdZ.ddsCaps.dwCaps = DDSCAPS_ZBUFFER;
		ddsdZ.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY; // require the z-buffer to be in video memory, otherwise it would be too slow

/// 	// used when software z-buffering is allowed (in which case the previous line should be disabled instead):
/// 	if(!win_back_buf_in_vid_mem || (d3d_device_desc->deviceGUID == IID_IDirect3DRGBDevice || d3d_device_desc->deviceGUID == IID_IDirect3DRefDevice))
/// 		ddsdZ.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		status = dd_main->CreateSurface(&ddsdZ, &d3d_zbuffer, NULL);
		if(SUCCEEDED(status)) {
			status = dd_back_buf->AddAttachedSurface(d3d_zbuffer);
			if(FAILED(status)) {
				if(win_err_debug)
					debug.Output("WinDX: Failed to attach z buffer.\n");
			} else
				d3d_zbuffer_enabled = true;
		} else {
			if(win_err_debug)
				debug.Output("Warning: Z-Buffering not supported on this computer.\n");
		}
	}
	// if z buffering hasn't been enabled at this point, depth sorting is used instead
	// (note that depth sorting is used even with z-buffering when drawing translucently)


	// make sure we assign a 3D device to the back buffer first,
	// or it probably won't get 3D hardware acceleration
	if(CreateDeviceForSurface(d3d_device, dd_back_buf) == false) {
		win_status = GAMEX_D3DSTART_FAILED;
		return false;
	}

	// set some render and texture state values, most of these may be the defaults, but just in case:

	d3d_device->SetTextureStageState(0,D3DTSS_ADDRESSU,D3DTADDRESS_CLAMP); // we don't want texture wrapping,
	d3d_device->SetTextureStageState(0,D3DTSS_ADDRESSV,D3DTADDRESS_CLAMP); // it doesn't work right on most image sizes anyway
	d3d_device->SetTextureStageState(0,D3DTSS_ALPHAOP, D3DTOP_MODULATE);
	d3d_device->SetTextureStageState(0,D3DTSS_MAGFILTER,D3DTFG_LINEAR);
	d3d_device->SetTextureStageState(0,D3DTSS_MINFILTER,D3DTFG_LINEAR);

	d3d_device->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE); // we have no D3D light sources, for now
	d3d_device->SetRenderState(D3DRENDERSTATE_SPECULARENABLE, FALSE); // don't need specular
	d3d_device->SetRenderState(D3DRENDERSTATE_COLORVERTEX, TRUE); // we want to allow per-vertex coloring
	d3d_device->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE); // we want to allow alpha blending
	d3d_device->SetRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE); // and dithering, except when temporarily disabled later
	d3d_device->SetRenderState(D3DRENDERSTATE_ANTIALIAS, FALSE); // set antialiasing off
	d3d_device->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW); // turn off culling of CW defined polygons if developer so desires
	d3d_device->SetRenderState(D3DRENDERSTATE_ZENABLE,FALSE); // the z buffer is only enabled while drawing in 3D
	d3d_device->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE,FALSE); 
	d3d_device->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
	d3d_device->SetRenderState(D3DRENDERSTATE_ALPHAREF, 248);
	d3d_device->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_GREATER);

	SetView(win_viewx1,win_viewy1,win_viewx2,win_viewy2); // now that D3D has been initialized set the view again for it

	draw_render_batch_device = NULL;

	if(d3d_zbuffer_enabled)
		d3d_device->Clear(0,NULL,D3DCLEAR_ZBUFFER,0,1.0f,0);

	return true;
}

bool WindowsDX::CreateDeviceForSurface (LPDIRECT3DDEVICE7 & device, LPDIRECTDRAWSURFACE7 & surface)
{
	if(win_request_software_only || FAILED(d3d_main->CreateDevice(IID_IDirect3DTnLHalDevice, surface, &device))) { // the best 3D hardware device
		if(win_request_software_only || FAILED(d3d_main->CreateDevice(IID_IDirect3DHALDevice, surface, &device))) { // second-best 3D hardware
			if(FAILED(d3d_main->CreateDevice(IID_IDirect3DMMXDevice, surface, &device))) { // best 3D software device
				// failing up until now is normal, although undesirable
				HRESULT status = d3d_main->CreateDevice(IID_IDirect3DRGBDevice, surface, &device); // backup 3D software
				if(FAILED(status)) {
					// that last try should never fail, something's very wrong if we get here
					DirectDrawError(status);
					Direct3DError(status);
					win_status = GAMEX_D3DSTART_FAILED;
					if(win_err_debug)
						debug.Output ("Direct3D: Device create failed.\n") ;
					return false; // should never get here; even if there's no 3D device, the D3DRGB device is software emulated
				}
			}
		}
	}
	return true;
}

void WindowsDX::DisplayLoadingWithValue (int percent)
{
	char str [64];
	sprintf(str, "Loading %d%%...", percent);
	ClearScreen() ;
	DrawGDIText(win_viewx2-120,win_viewy2-30,str) ;
	UpdateView() ;
	if(GetAsyncKeyState(VK_ESCAPE) & 0x8000)
		Quit(true);
}

// gets amount to shift a color component of the given mask
int WindowsDX::GetShift (int mask)
{
	if(mask == 0) return 0;
	int count;
	for (count=0; (mask & 1) == 0; count++)
		mask >>= 1;
	return count;
}

int WindowsDX::GetMax (int mask)
{
	if(mask == 0) return 0;
	int count, max; 
	for (count=0; (mask & 1) == 0; count++)
		mask >>= 1;
	for (max=0; mask > 0; max++)
		mask >>= 1;
	return max;
}


bool WindowsDX::InitDirectInput (void)
{
	// Create the DirectInput object:

	if(FAILED(DirectInputCreate(win_hinst, DIRECTINPUT_VERSION, &di_main, NULL)))
		return false;

	bool failed = false;

	// Create and set up the keyboard device:

	if(SUCCEEDED(di_main->CreateDevice(GUID_SysKeyboard, &di_keyboard, NULL))) {

		HRESULT status = DDERR_GENERIC;
		if(!win_request_no_auto_pause) {
			status = di_keyboard->SetCooperativeLevel(win_hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND | DISCL_NOWINKEY);
			if(SUCCEEDED(status))
				status = di_keyboard->SetDataFormat(&c_dfDIKeyboard);
			if(SUCCEEDED(status))
				status = di_keyboard->Acquire();
		}
		if(FAILED(status)) {
			status = di_keyboard->SetCooperativeLevel(win_hwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
			if(SUCCEEDED(status))
				status = di_keyboard->SetDataFormat(&c_dfDIKeyboard);
			if(SUCCEEDED(status))
				status = di_keyboard->Acquire();
		}
		if(FAILED(status))
			failed = true;
		else
			for(int k = 0 ; k < 255 ; k++)
				di_keystate[k] = di_keypressed[k] = 0;

	} else
		failed = true;

	if(!win_request_no_mouse_input) {

		// Create and set up the mouse device:

		if(SUCCEEDED(di_main->CreateDevice(GUID_SysMouse, &di_mouse, NULL))) {
			HRESULT status = DDERR_GENERIC;
			if(!win_request_no_auto_pause) {
				status = di_mouse->SetCooperativeLevel(win_hwnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);
				if(SUCCEEDED(status))
					status = di_mouse->SetDataFormat(&c_dfDIMouse);
				if(SUCCEEDED(status))
					status = di_mouse->Acquire();
			}
			if(FAILED(status)) {
				status = di_mouse->SetCooperativeLevel(win_hwnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
				if(SUCCEEDED(status))
					status = di_mouse->SetDataFormat(&c_dfDIMouse);
				if(SUCCEEDED(status))
					status = di_mouse->Acquire();
			}

			if(FAILED(status))
				failed = true;
		} else {
			failed = true;
			win_request_no_mouse_input = true;
		}
	} 

	UpdateInput(); // so the game can check for initial input inside GameInit() right after calling GameX.Initialize()

	return !failed;
}



bool WindowsDX::InitDirectSound (void)
{
	if(win_request_no_sound) return true;

	if(FAILED(DirectSoundCreate (NULL, &ds_main, NULL))) {
		win_status = GAMEX_SOUNDINIT_FAILED;
		ds_main = NULL;
		sound_support = false;
		return false;
	} else
	if(FAILED(ds_main->SetCooperativeLevel(win_hwnd, DSSCL_PRIORITY))) {
		ds_main->Release() ;
		win_status = GAMEX_SOUNDCOOP_FAILED;
		ds_main = NULL;
		sound_support = false;
		return false;
	}

	master_music_tempo = master_music_volume = 1.0f;

	sound_support = true;
	return true;
}

bool WindowsDX::InitDirectMusic (void)
{
	if(win_request_no_sound) return true;

	// set up DirectMusic and performance/ports:

	if(FAILED(CoCreateInstance(CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC,
								IID_IDirectMusicPerformance, (void**)&dm_performance)))
		return(false);

	if(FAILED(dm_performance->Init(NULL, (sound_support ? ds_main : NULL), win_hwnd)))
		return(false);

	if(FAILED(dm_performance->AddPort(NULL)))
		return(false);

	// prepare DirectMusic for loading MIDI files:

	if(FAILED(CoCreateInstance(CLSID_DirectMusicLoader, NULL, CLSCTX_INPROC,
								IID_IDirectMusicLoader, (void**)&dm_loader)))
		return(false);

	dm_musicx_index = 0;

	return true;
}

bool WindowsDX::TestDirectShow (void)
{
	IGraphBuilder * dsh_gb;
	HRESULT status = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC, IID_IGraphBuilder, (void **)&dsh_gb);
	if(FAILED(status)) {
		return false;
	} else {
		if (dsh_gb) 		dsh_gb->Release(),			 dsh_gb = NULL;
		return true;
	}
}



LPDIRECTDRAWSURFACE7 WindowsDX::CreateImageSurface (int width, int height, bool texture, int create_alpha, bool targetable)
{
	if(dd_main == NULL) {
		ReportProblem(GAMEX_NO_INIT) ; // loading image before GameX was initialized
	}

	LPDIRECTDRAWSURFACE7 surface; // the surface we'll create and return

	DDSURFACEDESC2 ddsd;
	ZeroMemory (&ddsd, sizeof(DDSURFACEDESC2)) ;
	ddsd.dwSize=sizeof(ddsd);

	DDPIXELFORMAT ddpf;
	ZeroMemory (&ddpf, sizeof(DDPIXELFORMAT)) ;
	ddpf.dwSize = sizeof(ddpf);

	ddsd.dwHeight=height;
	ddsd.dwWidth=width;
	ddsd.dwFlags=DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_CKSRCBLT;
	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue = 0; // specify black as transparent if drawn masked
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;
	if(texture && d3d_device != NULL) {
		ddsd.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
		D3DDEVICEDESC7 ddDesc;
		d3d_device->GetCaps(&ddDesc);
		if (ddDesc.deviceGUID == IID_IDirect3DHALDevice || ddDesc.deviceGUID == IID_IDirect3DTnLHalDevice) {
			ddsd.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
		}
		else {
			ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
		}

		if(targetable) { // && texture
			ddsd.ddsCaps.dwCaps |= DDSCAPS_3DDEVICE; // to allow the surface to request a 3D device
			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_HINTDYNAMIC; // could speed up images that are frequently drawn to
		} else {
			ddsd.ddsCaps.dwCaps2 |= DDSCAPS2_HINTSTATIC; // could speed up images that are never drawn to
		}

		if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_POW2) {
			// if the texture dimensions must be a power of 2,
			// increase the surface size accordingly (the change is transparent to the user and programmer)
			for(ddsd.dwWidth  = 1; (int)ddsd.dwWidth  < width ; ddsd.dwWidth  <<= 1);
			for(ddsd.dwHeight = 1; (int)ddsd.dwHeight < height; ddsd.dwHeight <<= 1);
		}
		if(ddDesc.dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_SQUAREONLY || targetable) { // for some reason DDSCAPS_3DDEVICE causes non-square textures to fail drawing
			// if the texture has to be square
			if(ddsd.dwWidth > ddsd.dwHeight) ddsd.dwHeight = ddsd.dwWidth;
			else							 ddsd.dwWidth  = ddsd.dwHeight;
		}
		int maxWidth = min((DWORD)d3d_max_tex_width, ddDesc.dwMaxTextureWidth);
		int maxHeight = min((DWORD)d3d_max_tex_height, ddDesc.dwMaxTextureHeight);
		if(ddsd.dwWidth  > (DWORD)maxWidth)  ddsd.dwWidth  = (DWORD)maxWidth;
		if(ddsd.dwHeight > (DWORD)maxHeight) ddsd.dwHeight = (DWORD)maxHeight;
		if(win_request_low_resolution) ddsd.dwWidth /= 2, ddsd.dwHeight /= 2;
		if(ddsd.dwWidth  < ddDesc.dwMinTextureWidth)  ddsd.dwWidth  = ddDesc.dwMinTextureWidth;
		if(ddsd.dwHeight < ddDesc.dwMinTextureHeight) ddsd.dwHeight = ddDesc.dwMinTextureHeight;
	} else {
		// if we're not making a 3D accelerated surface, just try for a 2D accelerated one
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	}

			if(create_alpha) {
				dd_back_buf->GetPixelFormat(&ddpf);
				ddpf.dwFlags = DDPF_RGB | DDPF_ALPHAPIXELS;
				ddpf.dwRGBBitCount	   = 32;
				ddpf.dwRGBAlphaBitMask = 0xFF000000;
				ddpf.dwRBitMask 	   = 0x00FF0000;
				ddpf.dwGBitMask 	   = 0x0000FF00;
				ddpf.dwBBitMask 	   = 0x000000FF;
				ddsd.dwFlags |= DDSD_PIXELFORMAT;
				ddsd.ddpfPixelFormat = ddpf;
			} 
			
	HRESULT status = dd_main->CreateSurface(&ddsd, &surface, NULL);
	if(FAILED(status)) {
		if(ddsd.dwFlags & DDSD_PIXELFORMAT) {
			ddsd.dwFlags ^= DDSD_PIXELFORMAT; // try again, this time accepting any format it wants to use
			status = dd_main->CreateSurface(&ddsd, &surface, NULL);
		}
		if(FAILED(status)) {
			DirectDrawError(status);
			if(win_err_debug)
				debug.Output("Failed to allocate image memory.\n");
			Quit(true);
			return NULL;
		}
	}

	if(win_status != GAMEX_NO_INIT) {
		surface->GetPixelFormat(&ddpf); // get the pixel format DirectDraw made the surface in,
		int size = (ddsd.dwWidth*ddsd.dwHeight*ddpf.dwRGBBitCount/8)/1024; if(size < 1) size = 1;
		if(win_log_debug) {
			if(GetStatus() == GAMEX_READY) {
				char sizeStr [8];
				sprintf(sizeStr, "%05d", size);
				for(int z = 0 ; sizeStr[z] == '0' ; z++) sizeStr[z] = ' ';

				debug.Output("%s KB, %dx%d%s -> %dx%dx%d\n",sizeStr,width,height,(create_alpha)?"+A":"", ddsd.dwWidth,ddsd.dwHeight,ddpf.dwRGBBitCount);
			}
		}
		MemoryRecordChange(size);
	}

	return surface;
}

void WindowsDX::ResetRandomSeed(void)
{
	srand((unsigned int)time(NULL)<<2);
}

void WindowsDX::SetRandomSeed(int seed)
{
	srand(seed);
}

bool WindowsDX::DirectXCleanUp (void)
{
	// DirectMusic MIDI songs:
	while(dm_musicx_index--)
		if(dm_musicx[dm_musicx_index] != NULL)
			dm_musicx[dm_musicx_index]->Release(),	dm_musicx[dm_musicx_index] = NULL;

	// DirectMusic
	if (dm_performance) dm_performance->CloseDown(), dm_performance->Release(), dm_performance = NULL;
	if (dm_loader)		dm_loader->Release(),		 dm_loader = NULL;
	
	// DirectSound
	{
		unsigned int size = (unsigned int)(sound_list.size());
		for(unsigned int i = 0 ; i < size ; ++i)
			sound_list[i]->Unload();
		sound_list.clear();
	}
	if (ds_main)		ds_main->Release(), 		ds_main = NULL;

	// DirectInput
	if (di_mouse)		di_mouse->Unacquire(),		di_mouse->Release(),		di_mouse = NULL;
	if (di_keyboard)	di_keyboard->Unacquire(),	di_keyboard->Release(), 	di_keyboard = NULL;
	if (di_main)		di_main->Release(), 		di_main = NULL;

	// Direct3D
	if (d3d_device) 	d3d_device->Release(),		d3d_device = NULL;	
	if (d3d_main)		d3d_main->Release(),		d3d_main = NULL;	

	// DirectDraw
	if (dd_back_buf)	dd_back_buf->Release(), 	dd_back_buf = NULL;
	if (dd_front_buf)	dd_front_buf->Release(),	dd_front_buf = NULL;	
	if (dd_main)		dd_main->Release(), 		dd_main = NULL;

	CoUninitialize();
	return true;
}

bool WindowsDX::CleanUp (void)
{
	if(win_log_debug) {
		debug.Output("-- Maximum Graphic/Sound memory in use: %d KB\n", memory_used_max);
		debug.Output("-- Graphic/Sound memory in use on exit: %d KB\n", memory_used_current);
	}
	if (DirectXCleanUp()==FALSE) {
		if(win_err_debug)
			debug.Output ("WinDX::CleanUp: Unable to properly shut down DirectX.\n") ;
		return false;
	}
	return true;
}
